/* -*- mode: C; mode: folding; fill-column: 70; -*- */
#define _XOPEN_SOURCE 600
#define _LARGEFILE64_SOURCE 1
#define _FILE_OFFSET_BITS 64

#include "stinger-atomics.h"
#include "stinger-utils.h"
#include "stinger.h"
#include "timer.h"
#include "xmalloc.h"

#define ACTI(k) (action[2*(k)])
#define ACTJ(k) (action[2*(k)+1])

static int64_t nv, ne, naction;
static int64_t * restrict off;
static int64_t * restrict from;
static int64_t * restrict ind;
static int64_t * restrict weight;
static int64_t * restrict action;

/* handles for I/O memory */
static int64_t * restrict graphmem;
static int64_t * restrict actionmem;

static char * initial_graph_name = INITIAL_GRAPH_NAME_DEFAULT;
static char * action_stream_name = ACTION_STREAM_NAME_DEFAULT;

static long batch_size = BATCH_SIZE_DEFAULT;
static long nbatch = NBATCH_DEFAULT;

static struct stinger * S;

static double * update_time_trace;


int
main (const int argc, char *argv[])
{
  parse_args (argc, argv, &initial_graph_name, &action_stream_name, &batch_size, &nbatch);
  STATS_INIT();

  load_graph_and_action_stream (initial_graph_name, &nv, &ne, (int64_t**)&off,
	      (int64_t**)&ind, (int64_t**)&weight, (int64_t**)&graphmem,
	      action_stream_name, &naction, (int64_t**)&action, (int64_t**)&actionmem);

  print_initial_graph_stats (nv, ne, batch_size, nbatch, naction);
  BATCH_SIZE_CHECK();

  update_time_trace = xmalloc (nbatch * sizeof(*update_time_trace));

  /* Convert to STINGER */
  tic ();
  S = stinger_new ();
  stinger_set_initial_edges (S, nv, 0, off, ind, weight, NULL, NULL, -2);
  PRINT_STAT_DOUBLE ("time_stinger", toc ());
  fflush(stdout);

  free(graphmem);

  /* Updates */
  int64_t ntrace = 0;
  if (batch_size == 1)
  {
    for (int64_t actk = 0; actk < nbatch; actk++, ntrace++) {
      tic();

      int64_t i = ACTI(actk);
      int64_t j = ACTJ(actk);

      int64_t changed;
      int64_t dir;
      int64_t incr = 0;
      int64_t num_affected = 0;
      int64_t delta_global_ntri;

      if (i < 0) {  /* This is a delete */
	i = -i-1;
	j = -j-1;
	dir = -1;

	assert (i >= 0);
	assert (i < nv);
	assert (j >= 0);
	assert (j < nv);

	changed = stinger_remove_edge_pair (S, 0, i, j);

      } else {  /* This is an insert */
	dir = 1;

	assert (i >= 0);
	assert (i < nv);
	assert (j >= 0);
	assert (j < nv);

	changed = stinger_insert_edge_pair (S, 0, i, j, 1, ntrace+1);

      }

      update_time_trace[ntrace] = toc();
    }
  } else { /* batching */

    int64_t * act = xmalloc (2 * 2 * batch_size * sizeof(*act));
    int64_t * insoff = xmalloc ((2 * 2 * batch_size + 1) * sizeof(*insoff));
    int64_t * deloff = xmalloc ((2 * 2 * batch_size + 1) * sizeof(*deloff));

    for (int64_t actno = 0; actno < nbatch * batch_size; actno += batch_size)
    {
      tic();

      const int64_t endact = (actno + batch_size > naction ? naction : actno + batch_size);
      int64_t N = stinger_sort_actions (endact - actno, &action[2*actno], insoff, deloff, act);

      stinger_remove_and_insert_batch (S, 0, actno+1, N, insoff, deloff, act);

      update_time_trace[ntrace] = toc();
      ntrace++;
    } /* End of batch */

    free (deloff);
    free (insoff);
    free (act);

  }

  /* Print the times */
  double time_updates = 0;
  for (int64_t k = 0; k < nbatch; k++) {
    time_updates += update_time_trace[k];
  }
  PRINT_STAT_DOUBLE ("time_updates", time_updates);
  PRINT_STAT_DOUBLE ("updates_per_sec", (nbatch * batch_size) / time_updates); 

  free(update_time_trace);
  stinger_free_all (S);
  free (actionmem);
  STATS_END();
}
