/* -*- mode: C; mode: folding; fill-column: 70; -*- */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <alloca.h>
#include <errno.h>
#include <assert.h>
#include <dirent.h>

#include "stinger.h"
#include "stinger-atomics.h"
#include "stinger-utils.h"
#include "xmalloc.h"

/* {{{ Internal utilities */


/** @brief Returns the out-degree of a vertex
 *
 *  @param S_ The STINGER data structure
 *  @param i_ Logical vertex ID
 *  @return Out-degree of vertex i_
 */
int64_t
stinger_outdegree (const struct stinger *S_, int64_t i_)
{
  return S_->LVA[i_].outDegree;
}

/** @brief Returns the out-degree of a vertex for a given edge type
 *
 *  @param S The STINGER data structure
 *  @param i Logical vertex ID
 *  @param type Edge type
 *  @return Out-degree of vertex i with type
 */
int64_t
stinger_typed_outdegree (const struct stinger * S, int64_t i, int64_t type) {
  int64_t out = 0;
  struct stinger_eb * current_eb = (S)->LVA[(i)].edges;
  while(current_eb) {
    if(current_eb->etype == type) {
      out += current_eb->numEdges;
    }
    current_eb = (current_eb->next);
  }
  return out;
}

/** @brief Calculate the largest active vertex ID
 *
 *  Finds the largest vertex ID whose in-degree is greater than zero and out-degree
 *  is greater than zero.
 *
 *  @param S The STINGER data structure
 *  @return Largest active vertex ID
 */
uint64_t
stinger_max_active_vertex(const struct stinger * S) {
  uint64_t out = 0;
  OMP("omp parallel") {
    uint64_t local_max = 0;
    OMP("omp for")
    for(uint64_t i = 0; i < STINGER_MAX_LVERTICES; i++) {
      if((stinger_indegree(S, i) > 0 || stinger_outdegree(S, i) > 0) && 
	i > local_max) {
	local_max = i;
      }
    }
    OMP("omp critical") {
      if(local_max > out)
	out = local_max;
    }
  }
  return out;
}

/** @brief Calculate the number of active vertices
 *
 *  Counts the number of vertices whose in-degree is greater than zero and
 *  out-degree is greater than zero.
 *
 *  @param S The STINGER data structure
 *  @return Number of active vertices
 */
uint64_t
stinger_num_active_vertices(const struct stinger * S) {
  uint64_t out = 0;
  OMP("omp parallel for reduction(+:out)")
  for(uint64_t i = 0; i < STINGER_MAX_LVERTICES; i++) {
    if(stinger_indegree(S, i) > 0 || stinger_outdegree(S, i) > 0) {
      out++;
    }
  }
  return out;
}

/** @brief Returns the in-degree of a vertex
 *
 *  @param S_ The STINGER data structure
 *  @param i_ Logical vertex ID
 *  @return In-degree of vertex i_
 */
int64_t
stinger_indegree (const struct stinger * S_, int64_t i_)
{
  return S_->LVA[i_].inDegree;
}

/** @brief Returns the vertex weight
 *
 *  @param S_ The STINGER data structure
 *  @param i_ Logical vertex ID
 *  @return Weight of vertex i_
 */
int64_t
stinger_vweight (const struct stinger * S_, int64_t i_)
{
  return S_->LVA[i_].weight;
}

/** @brief Sets the vertex weight
 *
 *  @param S_ The STINGER data structure
 *  @param i_ Logical vertex ID
 *  @param weight_ Vertex weight
 *  @return 1 on success
 */
int64_t
stinger_set_vweight (struct stinger * S_, int64_t i_, int64_t weight_)
{
  return (S_->LVA[i_].weight = weight_);
}

/** @brief Returns the vertex type
 *
 *  @param S_ The STINGER data structure
 *  @param i_ Logical vertex ID
 *  @return Type of vertex i_
 */
int64_t
stinger_vtype (const struct stinger * S_, int64_t i_)
{
  return S_->LVA[i_].vtype;
}

/** @brief Sets the vertex type
 *
 *  @param S_ The STINGER data structure
 *  @param i_ Logical vertex ID
 *  @param type_ Vertex type
 *  @return 1 on success
 */
int64_t
stinger_set_vtype (const struct stinger * S_, int64_t i_, int64_t type_)
{
  return (S_->LVA[i_].vtype = type_);
}

const struct stinger_eb *
stinger_edgeblocks (const struct stinger *S_, int64_t i_)
{
  return S_->LVA[i_].edges;
}

const struct stinger_eb *
stinger_next_eb (const struct stinger *G /*UNUSED*/,
                 const struct stinger_eb *eb_)
{
  return eb_->next;
}

int64_t
stinger_eb_type (const struct stinger_eb * eb_)
{
  return eb_->etype;
}

int
stinger_eb_high (const struct stinger_eb *eb_)
{
  return eb_->high;
}

int
stinger_eb_is_blank (const struct stinger_eb *eb_, int k_)
{
  return eb_->edges[k_].neighbor < 0;
}

int64_t
stinger_eb_adjvtx (const struct stinger_eb * eb_, int k_)
{
  return eb_->edges[k_].neighbor;
}

int64_t
stinger_eb_weight (const struct stinger_eb * eb_, int k_)
{
  return eb_->edges[k_].weight;
}

int64_t
stinger_eb_ts (const struct stinger_eb * eb_, int k_)
{
  return eb_->edges[k_].timeRecent;
}

int64_t
stinger_eb_first_ts (const struct stinger_eb * eb_, int k_)
{
  return eb_->edges[k_].timeFirst;
}

/**
* @brief Count the total number of edges in STINGER.
*
* @param S The STINGER data structure
*
* @return The number of edges in STINGER
*/
int64_t
stinger_total_edges (const struct stinger * S)
{
  uint64_t rtn = 0;
  for (uint64_t i = 0; i < STINGER_MAX_LVERTICES; i++) {
    rtn += stinger_outdegree (S, i);
  }
  return rtn;
}

void
stinger_print_eb(struct stinger_eb * eb) {
  printf(
    "EB VTX:  %ld\n"
    "  NEXT:    %p\n"
    "  TYPE:    %ld\n"
    "  NUMEDGS: %ld\n"
    "  HIGH:    %ld\n"
    "  SMTS:    %ld\n"
    "  LGTS:    %ld\n"
    "  EDGES:\n",
    eb->vertexID, eb->next, eb->etype, eb->numEdges, eb->high, eb->smallStamp, eb->largeStamp);
  uint64_t j = 0;
  for (; j < eb->high && j < STINGER_EDGEBLOCKSIZE; j++) {
    printf("    TO: %s%ld WGT: %ld TSF: %ld TSR: %ld\n", 
      eb->edges[j].neighbor < 0 ? "x " : "  ", eb->edges[j].neighbor < 0 ? ~(eb->edges[j].neighbor) : eb->edges[j].neighbor, 
      eb->edges[j].weight, eb->edges[j].timeFirst, eb->edges[j].timeRecent);
  }
  if(j < STINGER_EDGEBLOCKSIZE) {
    printf("  ABOVE HIGH:\n");
    for (; j < STINGER_EDGEBLOCKSIZE; j++) {
      printf("    TO: %ld WGT: %ld TSF: %ld TSR: %ld\n", 
	eb->edges[j].neighbor, eb->edges[j].weight, eb->edges[j].timeFirst, eb->edges[j].timeRecent);
    }
  }
}

/**
* @brief Checks the STINGER metadata for inconsistencies.
*
* @param S The STINGER data structure
* @param NV The total number of vertices
*
* @return 0 on success; failure otherwise
*/
int
stinger_consistency_check (struct stinger *S, uint64_t NV)
{
  uint64_t *inDegree = calloc (NV, sizeof (uint64_t));
  if (inDegree == NULL)
    return 1;

  // check blocks
  for (uint64_t i = 0; i < NV; i++) {
    uint64_t curOutDegree = 0;
    struct stinger_eb *curBlock = S->LVA[i].edges;
    while (curBlock) {
      if (curBlock->vertexID != i)
        return 2;
      if (curBlock->high > STINGER_EDGEBLOCKSIZE)
        return 3;

      int64_t numEdges = 0;
      int64_t smallStamp = INT64_MAX;
      int64_t largeStamp = INT64_MIN;

      uint64_t j = 0;
      for (; j < curBlock->high && j < STINGER_EDGEBLOCKSIZE; j++) {
        if (!stinger_eb_is_blank (curBlock, j)) {
          inDegree[stinger_eb_adjvtx (curBlock, j)]++;
          curOutDegree++;
          numEdges++;
          if (stinger_eb_ts (curBlock, j) < smallStamp)
            smallStamp = stinger_eb_ts (curBlock, j);
          if (stinger_eb_first_ts (curBlock, j) < smallStamp)
            smallStamp = stinger_eb_first_ts (curBlock, j);
          if (stinger_eb_ts (curBlock, j) > largeStamp)
            largeStamp = stinger_eb_ts (curBlock, j);
          if (stinger_eb_first_ts (curBlock, j) > largeStamp)
            largeStamp = stinger_eb_first_ts (curBlock, j);
        }
      }
      if (numEdges && numEdges != curBlock->numEdges)
        return 4;
      if (numEdges && largeStamp != curBlock->largeStamp)
        return 5;
      if (numEdges && smallStamp != curBlock->smallStamp)
        return 6;
      for (; j < STINGER_EDGEBLOCKSIZE; j++) {
        if (!(stinger_eb_is_blank (curBlock, j) ||
              (stinger_eb_adjvtx (curBlock, j) == 0
               && stinger_eb_weight (curBlock, j) == 0
               && stinger_eb_ts (curBlock, j) == 0
               && stinger_eb_first_ts (curBlock, j) == 0)))
          return 7;
      }
      curBlock = curBlock->next;
    }

    if (curOutDegree != S->LVA[i].outDegree)
      return 8;
  }

  for (uint64_t i = 0; i < NV; i++) {
    if (inDegree[i] != S->LVA[i].inDegree)
      return 9;
  }

  free (inDegree);

  return 0;
}


/* }}} */

/* {{{ Internal structure parameters */

#define INITIAL_EBLK_PER_ETYPE 8192u
#define INIT_N_EBPOOL_PTRS (1u<<14)
#define EBPOOL_SIZE (STINGER_MAX_LVERTICES*2)

/* }}} */

/* {{{ STINGER default parameters */
#define DEFAULT_VTYPE 0
#define DEFAULT_VWEIGHT 1
/* }}} */

/* {{{ Edge block pool */

static size_t which_ebpool, cur_ebpool_tail;
static struct stinger_eb **ebpool = NULL;

MTA ("mta inline")
static void init_ebpool (void)
{
  OMP ("omp critical (stinger__init_ebpool)") {
    struct stinger_eb **new_ebpool;
#if defined(__MTA__)
    new_ebpool = readfe (&ebpool);
    if (new_ebpool) {
      writeef (&ebpool, new_ebpool);
      return;
    }
#else
    if (!ebpool)                /* All because OpenMP can't handle return... */
#endif
      {
        new_ebpool = xcalloc (INIT_N_EBPOOL_PTRS, sizeof (*ebpool));
        new_ebpool[0] = xmalloc (EBPOOL_SIZE * sizeof (struct stinger_eb));
        which_ebpool = 0;
        cur_ebpool_tail = 0;
#if !defined(__MTA__)
        ebpool = new_ebpool;
#else
        writeef (&ebpool, new_ebpool);
#endif
      }
  }
}

static void
free_ebpool (void)
{
  for (uint64_t i = 0; i <= which_ebpool; ++i)
    free (ebpool[i]);
  free (ebpool);
  ebpool = NULL;
}

static void
blast_ebpool (void)
{
  which_ebpool = 0;
  cur_ebpool_tail = 0;
}

MTA ("mta expect parallel context") MTA ("mta inline")
static void
get_from_ebpool (struct stinger_eb **out, size_t k)
{
  size_t wp0, ebt0, n0 = 0;
  size_t wp1, ebt1, n1 = 0;
  {
    struct stinger_eb *restrict ebp;
    wp0 = which_ebpool;
    ebt0 = stinger_size_fetch_add (&cur_ebpool_tail, k);
    ebp = &ebpool[wp0][ebt0];
    if (ebt0 + k >= EBPOOL_SIZE) {
      fprintf (stderr, "XXX: eb pool exhausted\n");
      abort ();
    }

    OMP ("omp parallel for")
      MTA ("mta assert nodep")
      MTASTREAMS ()MTA ("mta block schedule")
      for (size_t ki = 0; ki < k; ++ki)
        out[ki] = &ebp[ki];
  }
}

/* }}} */

/* {{{ Allocating and tearing down */

MTA ("mta inline")
void stinger_init (void)
{
  if (!ebpool)
    init_ebpool ();
}


/** @brief Create a new STINGER data structure.
 *
 *  Allocates memory for a STINGER data structure.  If this is the first STINGER
 *  to be allocated, it also initializes the edge block pool.  Edge blocks are
 *  allocated and assigned for each value less than STINGER_NUMETYPES.
 *
 *  @return Pointer to struct stinger
 */
MTA ("mta inline")
struct stinger *stinger_new (void)
{
  struct stinger *G = xcalloc (1, sizeof (*G));
  size_t i;

  if (!ebpool)
    stinger_init ();

  G->LVA = xcalloc (STINGER_MAX_LVERTICES, sizeof (G->LVA[0]));
  G->LVASize = STINGER_MAX_LVERTICES;

#if STINGER_NUMETYPES > 1
  OMP ("omp parallel for")
    MTA ("mta assert parallel")
    MTASTREAMS ()
    for (i = 0; i < STINGER_NUMETYPES; ++i) {
      struct stinger_eb **blocks;
      blocks = xcalloc (INITIAL_EBLK_PER_ETYPE, sizeof (*blocks));
      G->ETA[i].blocks = blocks;
      G->ETA[i].length = INITIAL_EBLK_PER_ETYPE;
    }
#else
  G->ETA[0].blocks =
    xcalloc (INITIAL_EBLK_PER_ETYPE, sizeof (struct stinger_eb *));
  G->ETA[0].length = INITIAL_EBLK_PER_ETYPE;
#endif

  return G;
}

/*
 * TODO - in progress
 struct stinger * 
 stinger_copy(struct stinger * S, uint64_t nv) {
 struct stinger * G = xcalloc(1, sizeof(*G));

 if (!ebpool) stinger_init ();

 G->LVA = xcalloc (STINGER_MAX_LVERTICES, sizeof (G->LVA[0]));
 G->LVASize = STINGER_MAX_LVERTICES;
 #if STINGER_NUMETYPES > 1
 OMP("omp parallel for")
 MTA("mta assert parallel")
 MTASTREAMS()
 for (i = 0; i < STINGER_NUMETYPES; ++i) {
 G->ETA[i].blocks = xcalloc (S->ETA[i].length, sizeof (struct stinger_eb*));
 G->ETA[i].length = S->ETA[i].length;
 G->ETA[i].high = S->ETA[i].high;
 }
 #else
 G->ETA[0].blocks = xcalloc (S->ETA[0].length, sizeof (struct stinger_eb*));
 G->ETA[0].length = S->ETA[0].length;
 G->ETA[0].high = S->ETA[0].high;
 #endif

 return G;
 }
*/

/** @brief Free memory allocated to a particular STINGER instance.
 *
 *  Frees the ETA pointers for each edge type, the LVA, and the struct stinger
 *  itself.  Does not actually free any edge blocks, as there may be other
 *  active STINGER instances.
 *
 *  @param S The STINGER data structure
 *  @return NULL on success
 */
struct stinger *
stinger_free (struct stinger *S)
{
  size_t i;
  if (!S)
    return S;
  for (i = 0; i < STINGER_NUMETYPES; ++i) {
    /* Don't free the blocks themselves.  They live
       in the unmanaged pool. */
    free (S->ETA[i].blocks);
  }
  free (S->LVA);
  free (S);
  free_ebpool ();
  return NULL;
}

/** @brief Free the STINGER data structure and all edge blocks.
 *
 *  Free memory allocated to the specified STINGER data structure.  Also frees
 *  the STINGER edge block pool, effectively ending all STINGER operations globally.
 *  Only call this function if you are done using STINGER entirely.
 *
 *  @param S The STINGER data structure
 *  @return NULL on success
 */
struct stinger *
stinger_free_all (struct stinger *S)
{
  struct stinger *out;
  out = stinger_free (S);
  blast_ebpool ();
  return out;
}

MTA ("mta inline")
static void zero_mem (void *x, const size_t sz)
{
#if !defined(__MTA__)
  memset (x, 0, sz);
#else
  /* No, really.  They couldn't be bothered to implement memset intelligently. */
  unsigned char *restrict t = (unsigned char *) x;
  MTA ("mta block schedule") MTASTREAMS ()
    for (size_t k = 0; k < sz; ++k)
      t[k] = 0;
#endif
}

MTA ("mta expect parallel context")
static struct stinger_eb *new_eb (int64_t etype, int64_t from)
{
  size_t k;
  struct stinger_eb *out = NULL;
  //out = xcalloc (1, sizeof (*out));
  get_from_ebpool (&out, 1);
  assert (out);
  zero_mem (out, sizeof (*out));
  out->etype = etype;
  out->vertexID = from;
  out->smallStamp = INT64_MAX;
  out->largeStamp = INT64_MIN;
  return out;
}

MTA ("mta expect parallel context")
static void
new_ebs (struct stinger_eb **out, size_t neb, int64_t etype,
         int64_t from)
{
  if (neb < 1)
    return;
  get_from_ebpool (out, neb);

  OMP ("omp parallel for")
    //MTA("mta assert nodep")
    MTASTREAMS ()MTA ("mta block schedule")
    //MTA("mta parallel single processor")
    for (size_t i = 0; i < neb; ++i) {
      zero_mem (out[i], sizeof (*out[i]));
      out[i]->etype = etype;
      out[i]->vertexID = from;
      out[i]->smallStamp = INT64_MAX;
      out[i]->largeStamp = INT64_MIN;
    }
}

MTA ("mta expect serial context")
static void
new_blk_ebs (struct stinger_eb **out, const struct stinger *restrict G,
             const int64_t nvtx, const size_t * restrict blkoff,
             const int64_t etype)
{
  size_t neb;
  if (nvtx < 1)
    return;
  neb = blkoff[nvtx];
  get_from_ebpool (out, neb);

  OMP ("omp parallel for")
    MTA ("mta assert nodep")
    MTASTREAMS ()MTA ("mta block schedule")
    for (size_t k = 0; k < neb; ++k) {
      zero_mem (out[k], sizeof (*out[k]));
      out[k]->etype = etype;
      out[k]->smallStamp = INT64_MAX;
    }

  OMP ("omp parallel for")
    MTA ("mta assert nodep")
    MTASTREAMS ()MTA ("mta interleave schedule")
    for (int64_t v = 0; v < nvtx; ++v) {
      const int64_t from = v;
      const size_t blkend = blkoff[v + 1];
      MTA ("mta assert nodep")
        for (size_t k = blkoff[v]; k < blkend; ++k)
          out[k]->vertexID = from;
      if (blkend)
        MTA ("mta assert nodep")
          for (size_t k = blkoff[v]; k < blkend - 1; ++k)
            out[k]->next = out[k + 1];
    }
}

/* }}} */

MTA ("mta inline")
static void
push_ebs (struct stinger *G, size_t neb,
          struct stinger_eb *restrict * eb)
/* XXX: Reallocing here will nuke current readers *AND* writers.
   A DCAS could save the writers, but we need RCU-like machinery
   for readers. */
{
  int64_t etype, place;
  assert (G);
  assert (eb);

  if (!neb)
    return;

  etype = eb[0]->etype;
  assert (etype >= 0);
  assert (etype < STINGER_NUMETYPES);

  place = stinger_int64_fetch_add (&(G->ETA[etype].high), neb);

  OMP ("omp critical (stinger__push_eb)") {
    struct stinger_eb **blocks;
#if !defined(__MTA__) || 1      // critical section basically off on XMT
    blocks = G->ETA[etype].blocks;
#else
    /* Stop new readers, not active ones: */
    blocks = readfe (&G->ETA[etype].blocks);
#endif
    if (place + neb - 1 >= G->ETA[etype].length) {
      size_t newlen;
      struct stinger_eb **oldblocks;

      oldblocks = blocks;
      newlen = 2 * (place + neb - 1);

      blocks = xrealloc (oldblocks, newlen * sizeof (*blocks));
      zero_mem (&blocks[G->ETA[etype].length],
                (newlen - G->ETA[etype].length) * sizeof (*blocks));
      G->ETA[etype].length = newlen;
    }

    MTA ("mta assert nodep")
      for (int64_t k = 0; k < neb; ++k)
        blocks[place + k] = eb[k];
#if !defined(__MTA__) || 1
    G->ETA[etype].blocks = blocks;
#else
    writeef (&G->ETA[etype].blocks, blocks);
#endif
  }
}

struct curs
{
  struct stinger_eb *eb, **loc;
};

MTA ("mta inline")
static struct curs etype_begin (struct stinger_vb *v, int etype)
{
  struct curs out;
  assert (v);
  out.eb = v->edges;
  out.loc = &v->edges;
  while (out.eb && out.eb->etype < etype) {
    out.loc = &out.eb->next;
    out.eb = out.eb->next;
  }
  return out;
}

MTA ("mta inline")
static void
update_edge_data (struct stinger * S, struct stinger_eb *eb,
                  uint64_t index, int64_t neighbor, int64_t weight, int64_t ts)
{
  struct stinger_edge * e = eb->edges + index;

  /* new edge */
  if (e->neighbor < 0 
      || (e->neighbor == 0 && e->weight == 0 
	&& e->timeFirst == 0 && e->timeRecent == 0)) {
    e->neighbor = neighbor;
    e->timeFirst = ts;

    /* only edge in block */
    if(stinger_int64_fetch_add(&eb->numEdges, 1) == 0) {
      eb->smallStamp = ts;
      eb->largeStamp = ts;
    }

    /* register new edge */
    stinger_int64_fetch_add (&(S->LVA[eb->vertexID].outDegree), 1);
    stinger_int64_fetch_add (&(S->LVA[neighbor].inDegree), 1);
  }

  e->weight = weight;
  e->timeRecent = ts;

  /* maintain metadata */
  if (ts < eb->smallStamp)
    eb->smallStamp = ts;
  if (ts > eb->largeStamp)
    eb->largeStamp = ts;
  if (index >= eb->high)
    eb->high = index +1;
}

MTA ("mta inline")
static void
remove_edge(struct stinger * S, struct stinger_eb *eb, uint64_t index) {
  struct stinger_edge * e = eb->edges + index;

  /* already deleted */
  if(e->neighbor < 0)
    return;

  /* unregister edge */
  stinger_int64_fetch_add (&(S->LVA[eb->vertexID].outDegree), -1);
  stinger_int64_fetch_add (&(S->LVA[e->neighbor].inDegree), -1);

  e->neighbor = ~e->neighbor;

  /* maintain metadata */
  stinger_int64_fetch_add(&eb->numEdges, -1);
  if (index == eb->high+1)
    eb->high = index;

  if (e->timeFirst == eb->smallStamp || e->timeRecent == eb->smallStamp) {
    eb->smallStamp = INT64_MAX;
    for(uint64_t i = 0; i < eb->high; i++) {
      if(!stinger_eb_is_blank(eb, i)) {
	if(eb->edges[i].timeFirst < eb->smallStamp)
	  eb->smallStamp = eb->edges[i].timeFirst;
	if(eb->edges[i].timeRecent < eb->smallStamp)
	  eb->smallStamp = eb->edges[i].timeRecent;
      }
    }
  }
  if (e->timeFirst == eb->largeStamp || e->timeRecent == eb->largeStamp) {
    eb->largeStamp = INT64_MIN;
    for(uint64_t i = 0; i < eb->high; i++) {
      if(!stinger_eb_is_blank(eb, i)) {
	if(eb->edges[i].timeFirst > eb->largeStamp)
	  eb->largeStamp = eb->edges[i].timeFirst;
	if(eb->edges[i].timeRecent > eb->largeStamp)
	  eb->largeStamp = eb->edges[i].timeRecent;
      }
    }
  }
}

MTA ("mta inline")
static void
incr_edge_data (struct stinger_eb *eb,
                struct stinger_edge *e, int64_t weight, int64_t ts)
{
  if (UNLIKELY
      (e->neighbor < 0
       || (e->neighbor == 0 && e->weight == 0 && e->timeFirst == 0)
       || e->timeFirst > ts))
    /* Catch initial input or replaying an old stream. */
    e->timeFirst = ts;
  stinger_int64_fetch_add (&e->weight, weight);
  e->timeRecent = ts;
  if (ts < eb->smallStamp)
    eb->smallStamp = ts;
  if (ts > eb->largeStamp)
    eb->largeStamp = ts;
}

MTA ("mta inline") MTA ("mta serial")
static int
insert_active_edge (struct stinger *G,
                    int64_t type, int64_t v, int64_t w,
                    int64_t weight, int64_t timestamp)
{
  /* Do *NOT* call this concurrently with the same from value, even
  with different types. */
  struct curs curs;
  struct stinger_eb *tmp, *has_slot;
  size_t kslot;

  curs = etype_begin (&G->LVA[v], type);
  /* XXX: Racy throughout. */
  /*
  Possibilities:
  1: Edge already exists and only needs updated.
  2: Edge does not exist, fits in an existing block.
  3: Edge does not exist, needs a new block.
  */

  /* 1: Check if the edge already exists. */
  /*     (also for 2, record an empty slot.) */
  has_slot = NULL;
  kslot = STINGER_EDGEBLOCKSIZE;
  for (tmp = curs.eb; tmp && type == tmp->etype; tmp = tmp->next) {
    size_t k, endk;
    size_t found_k = STINGER_EDGEBLOCKSIZE,
    found_slot = STINGER_EDGEBLOCKSIZE;
    endk = tmp->high;
    if (!has_slot && endk < STINGER_EDGEBLOCKSIZE) {
      /* Prefer pushing on the end. */
      has_slot = tmp;
      kslot = endk;
    }
#if !defined(__MTA__)
    OMP ("omp parallel for")
    for (k = 0; k < endk; ++k) {
      if (w == tmp->edges[k].neighbor)
	found_k = k;
      else if (!has_slot && tmp->edges[k].neighbor < 0)
	/* Don't care *which* slot is found... */
	found_slot = k;
    }
#else
    /* Fusing the two loops requires an mta assert parallel, and that
    causes found_k to point at an existing edge for some reason.
    */
    MTA ("mta block schedule") MTASTREAMS ()
    for (k = 0; k < endk; ++k) {
      if (w == tmp->edges[k].neighbor)
	found_k = k;
    }
    if (!has_slot && found_k >= STINGER_EDGEBLOCKSIZE) {
      for (k = 0; k < endk; ++k) {
	if (tmp->edges[k].neighbor < 0)
	  /* Don't care *which* slot is found... */
	  found_slot = k;
      }
    }
#endif
    if (found_k < STINGER_EDGEBLOCKSIZE) {
      update_edge_data (G, tmp, found_k, w, weight, timestamp);
      return 0;
    } else if (found_slot < STINGER_EDGEBLOCKSIZE) {
      has_slot = tmp;
      kslot = found_slot;
    }
  }

  /* 2: The edge isn't already there.  Check for an empty slot. */
  if (has_slot) {
    assert (kslot < STINGER_EDGEBLOCKSIZE);
    assert (kslot >= has_slot->high || has_slot->edges[kslot].neighbor < 0);
    MTA ("mta fence") OMP ("omp flush") 
    update_edge_data (G, has_slot, kslot, w, weight, timestamp);
    return 1;
  }

  /* 3: Needs a new block to be inserted at curs.loc. */
  tmp = new_eb (type, v);
  update_edge_data (G, tmp, 0, w, weight, timestamp);

  tmp->next = curs.eb;
  MTA ("mta fence") OMP ("omp flush") * curs.loc = tmp;
  push_ebs (G, 1, &tmp);
  return 1;
}

MTA ("mta inline")
static int
incr_active_edge (struct stinger *G,
                  int64_t type, int64_t v, int64_t w,
                  int64_t weight, int64_t timestamp)
{
  /* Do *NOT* call this concurrently with the same from value, even
     with different types. */
  struct curs curs;
  struct stinger_eb *tmp, *has_slot;
  size_t kslot;

  curs = etype_begin (&G->LVA[v], type);
  /* XXX: Racy throughout. */
  /*
    Possibilities:
    1: Edge already exists and only needs updated.
    2: Edge does not exist, fits in an existing block.
    3: Edge does not exist, needs a new block.
  */

  /* 1: Check if the edge already exists. */
  /*     (also for 2, record an empty slot.) */
  has_slot = NULL;
  kslot = STINGER_EDGEBLOCKSIZE;
  for (tmp = curs.eb; tmp && type == tmp->etype; tmp = tmp->next) {
    size_t k, endk;
    size_t found_k = STINGER_EDGEBLOCKSIZE,
      found_slot = STINGER_EDGEBLOCKSIZE;
    endk = tmp->high;
    if (!has_slot && endk < STINGER_EDGEBLOCKSIZE) {
      /* Prefer pushing on the end. */
      has_slot = tmp;
      kslot = endk;
    }
#if !defined(__MTA__)
    OMP ("omp parallel for")
      MTA ("mta assert nodep")
      MTA ("mta block schedule") MTASTREAMS ()
      for (k = 0; k < endk; ++k) {
        if (w == tmp->edges[k].neighbor)
          found_k = k;
        else if (!has_slot && tmp->edges[k].neighbor < 0)
          /* Don't care *which* slot is found... */
          found_slot = k;
      }
#else
    /* Fusing the two loops requires an mta assert parallel, and that
       causes found_k to point at an existing edge for some reason.
    */
    MTA ("mta block schedule") MTASTREAMS ()
      for (k = 0; k < endk; ++k)
        if (w == tmp->edges[k].neighbor)
          found_k = k;
    if (!has_slot && found_k >= STINGER_EDGEBLOCKSIZE)
      for (k = 0; k < endk; ++k)
        if (tmp->edges[k].neighbor < 0)
          /* Don't care *which* slot is found... */
          found_slot = k;
#endif
    if (found_k < STINGER_EDGEBLOCKSIZE) {
      assert (tmp->edges[found_k].neighbor == w);
      incr_edge_data (tmp, &tmp->edges[found_k], weight, timestamp);
      return 0;
    } else if (found_slot < STINGER_EDGEBLOCKSIZE) {
      has_slot = tmp;
      kslot = found_slot;
    }
  }

  /* 2: The edge isn't already there.  Check for an empty slot. */
  if (has_slot) {
    assert (kslot < STINGER_EDGEBLOCKSIZE);
    assert (kslot >= has_slot->high || has_slot->edges[kslot].neighbor < 0);
    MTA ("mta fence") OMP ("omp flush") 
    update_edge_data (G, has_slot, kslot, w, weight, timestamp);
  } else {
    /* 3: Needs a new block to be inserted at curs.loc. */
    tmp = new_eb (type, v);
    update_edge_data (G, tmp, 0, w, weight, timestamp);

    tmp->next = curs.eb;
    MTA ("mta fence") OMP ("omp flush") * curs.loc = tmp;
    push_ebs (G, 1, &tmp);
  }
  return 1;
}

/** @brief Insert a directed edge.
 *
 *  Inserts a typed, directed edge.  First timestamp is set, if the edge is
 *  new.  Recent timestamp is updated.  Weight is set to specified value regardless.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @param weight Edge weight
 *  @param timestamp Edge timestamp
 *  @return 1
 */
MTA ("mta inline")
int
stinger_insert_edge (struct stinger *G,
                     int64_t type, int64_t from, int64_t to,
                     int64_t weight, int64_t timestamp)
{
  /* Do *NOT* call this concurrently with the same from value, even
     with different types. */
  STINGERASSERTS ();

  return insert_active_edge (G, type, from, to, weight, timestamp);
}

/** @brief Increments a directed edge.
 *
 *  Increments the weight of a typed, directed edge.
 *  Recent timestamp is updated.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @param weight Edge weight
 *  @param timestamp Edge timestamp
 *  @return 1
 */
MTA ("mta inline")
int
stinger_incr_edge (struct stinger *G,
                   int64_t type, int64_t from, int64_t to,
                   int64_t weight, int64_t timestamp)
{
  /* Do *NOT* call this concurrently with the same from value, even
     with different types. */
  STINGERASSERTS ();

  return incr_active_edge (G, type, from, to, weight, timestamp);
}

/** @brief Insert an undirected edge.
 *
 *  Inserts a typed, undirected edge.  First timestamp is set, if the edge is
 *  new.  Recent timestamp is updated.  Weight is set to specified value regardless.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @param weight Edge weight
 *  @param timestamp Edge timestamp
 *  @return 1
 */
MTA ("mta inline")
int
stinger_insert_edge_pair (struct stinger *G,
                          int64_t type, int64_t from, int64_t to,
                          int64_t weight, int64_t timestamp)
{
  int64_t v, w;

  assert (G);

  v = from;
  w = to;
  assert (v >= 0);
  assert (w >= 0);
  assert (v < STINGER_MAX_LVERTICES);
  assert (w < STINGER_MAX_LVERTICES);

  if (v != w) {
#if defined(__MTA__)
    future int ft$, tf$;
    future ft$ (G, type, v, w, weight, timestamp) {
      return insert_active_edge (G, type, v, w, weight, timestamp);
    }
    future tf$ (G, type, v, w, weight, timestamp) {
      return insert_active_edge (G, type, w, v, weight, timestamp);
    }
    return tf$ + ft$;
#else
    int out = 0;
    OMP ("omp parallel sections reduction(||:out)") {
      OMP ("omp section")
        out = insert_active_edge (G, type, v, w, weight, timestamp);

      OMP ("omp section")
        out = insert_active_edge (G, type, w, v, weight, timestamp);
    }
    return out;
#endif
  } else
    return insert_active_edge (G, type, v, w, weight, timestamp);
}

/** @brief Increments an undirected edge.
 *
 *  Increments the weight of a typed, undirected edge.
 *  Recent timestamp is updated.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @param weight Edge weight
 *  @param timestamp Edge timestamp
 *  @return 1
 */
MTA ("mta inline")
int
stinger_incr_edge_pair (struct stinger *G,
                        int64_t type, int64_t from, int64_t to,
                        int64_t weight, int64_t timestamp)
{
  STINGERASSERTS ();
  if (from != to) {
#if defined(__MTA__)
    future int ft$, tf$;
    future ft$ (G, type, from, to, weight, timestamp) {
      return incr_active_edge (G, type, from, to, weight, timestamp);
    }
    future tf$ (G, type, from, to, weight, timestamp) {
      return incr_active_edge (G, type, to, from, weight, timestamp);
    }
    /* This dies if you use && rather than +. */
    return tf$ + ft$;
#else
    int out = 0;
    OMP ("omp parallel sections reduction(||:out)") {
      OMP ("omp section")
        out = incr_active_edge (G, type, from, to, weight, timestamp);

      OMP ("omp section")
        out = incr_active_edge (G, type, to, from, weight, timestamp);
    }
    return out;
#endif
  } else
    return incr_active_edge (G, type, from, to, weight, timestamp);
}

MTA ("mta inline")
static int
remove_active_edge (struct stinger *G,
                    int64_t type, int64_t v, int64_t w)
{
  /* Do *NOT* call this concurrently with the same from value, even
     with different types. */
  struct curs curs;
  struct stinger_eb *tmp;

  curs = etype_begin (&G->LVA[v], type);
  /* XXX: Racy throughout. */
  /*
    Difference from insertion: If the edge doesn't exist...
  */

  /* Look for the edge. */
  for (tmp = curs.eb; tmp && type == tmp->etype; tmp = tmp->next) {
    size_t k, endk;
    size_t found_k = STINGER_EDGEBLOCKSIZE;
    endk = tmp->high;
    OMP ("omp parallel for")
      MTA ("mta assert nodep")
      MTA ("mta block schedule") MTASTREAMS ()
      for (k = 0; k < endk; ++k)
        if (w == tmp->edges[k].neighbor)
          found_k = k;
    if (found_k < STINGER_EDGEBLOCKSIZE) {
      remove_edge(G, tmp, found_k);
      return 1;
    }
  }
  return 0;
}

/** @brief Removes a directed edge.
 *
 *  Remove a typed, directed edge.
 *  Note: Do not call this function concurrently with the same source vertex,
 *  even for different edge types.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @return 1 on success, 0 if the edge is not found.
 */
MTA ("mta inline")
int
stinger_remove_edge (struct stinger *G,
                     int64_t type, int64_t from, int64_t to)
{
  /* Do *NOT* call this concurrently with the same from value, even
     with different types. */
  STINGERASSERTS ();

  return remove_active_edge (G, type, from, to);
}

/** @brief Removes an undirected edge.
 *
 *  Remove a typed, undirected edge.
 *  Note: Do not call this function concurrently with the same source vertex,
 *  even for different edge types.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @return 1 on success, 0 if the edge is not found.
 */
MTA ("mta inline")
int
stinger_remove_edge_pair (struct stinger *G,
                          int64_t type, int64_t from, int64_t to)
{
  /* Do *NOT* call this concurrently with the same from value, even
     with different types. */
  int64_t v, w;

  assert (G);

  /* This stil activates the vertices. */
  v = from;
  w = to;
  assert (v >= 0);
  assert (w >= 0);
  assert (v < STINGER_MAX_LVERTICES);
  assert (w < STINGER_MAX_LVERTICES);

  if (v != w) {
#if defined(__MTA__)
    future int ft$, tf$;
    future ft$ (G, type, v, w) {
      return remove_active_edge (G, type, v, w);
    }
    future tf$ (G, type, v, w) {
      return remove_active_edge (G, type, w, v);
    }
    return tf$ + ft$;
#else
    int out = 0;
    OMP ("omp parallel sections reduction(||:out)") {
      OMP ("omp section") out = remove_active_edge (G, type, v, w);

      OMP ("omp section") out = remove_active_edge (G, type, w, v);
    }
    return out;
#endif
  } else
    return remove_active_edge (G, type, v, w);
}

static int
i64_cmp (const void *a, const void *b)
{
  const int64_t va = *(const int64_t *) a;
  const int64_t vb = *(const int64_t *) b;
  return va - vb;
}

#define SEQLEN 4

MTA ("mta inline") MTA ("mta expect parallel context")
int64_t
find_in_sorted (const int64_t tofind,
                const int64_t N, const int64_t * restrict ary)
{
  int64_t out = -1;
  if (N <= 0)
    return -1;

  int64_t top = N - 1, bot = 0;

  if (tofind == ary[bot])
    return bot;
  if (tofind == ary[top])
    return top;
  while (top - bot + 1 > SEQLEN) {
    int64_t mid = bot + ((top - bot) / 2);
    if (tofind == ary[mid])
      return mid;
    if (tofind < ary[mid])
      top = mid;
    else
      bot = mid;
  }
  for (; bot <= top; ++bot)
    if (tofind == ary[bot])
      return bot;
  return -1;
}

/** @brief Remove and insert edges incident on a common source vertex.
 *
 *  For a given source vertex and edge type, removes a list of edges given in
 *  an array.  Then inserts a list of edges given in a second array with weight
 *  and a timestamp.
 *  Note: Do not call this function concurrently with the same source vertex,
 *  even for different edge types.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param v Source vertex ID
 *  @param nremove Number of edges to remove
 *  @param remove Edge adjacencies to remove
 *  @param ninsert Number of edges to insert
 *  @param insert Edge adjacencies to insert
 *  @param weight Edge adjacency weights to insert
 *  @param timestamp Timestamp for all edge insertions
 *  @return 0 on success; number of failures otherwise
 */
MTA ("mta parallel off") MTA ("mta expect parallel context")
static int
remove_and_insert_active_edges (struct stinger *G,
                                int64_t type, int64_t v,
                                int64_t nremove,
                                int64_t * restrict remove,
                                int64_t ninsert,
                                int64_t * restrict insert,
                                int64_t * restrict weight,
                                int64_t timestamp)
{
  /* Do *NOT* call this concurrently with the same from value, even
     with different types. */
  /*
    First pass:
    - Remove vertices.
    - Scan for insertion slots.
    - Scan for existing vertices being inserted.
    Second:
    - Insert into saved locations.
  */

  int64_t nrem = 0;
  struct curs curs;
  struct stinger_eb *tmp, **prev_loc;
  struct stinger_eb **has_slot = NULL;
  int *kslot = NULL;
  int64_t nslot = 0, ninsert_remaining = ninsert;

  if (1 == nremove) {
    if (remove[0] >= 0)
      nrem = remove_active_edge (G, type, v, remove[0]);
    nremove = 0;
  }
  if (1 == ninsert) {
    if (insert[0] >= 0)
      nrem += insert_active_edge (G, type, v, insert[0],
                                  (weight ? weight[0] : 1), timestamp);
    ninsert = 0;
  }

  /* Sort for binary search */
  if (nremove) {
    qsort (remove, nremove, sizeof (*remove), i64_cmp);
    while (nremove && *remove < 0) {
      ++remove;
      --nremove;
    }
  }
  if (ninsert) {
    has_slot = xmalloc (ninsert * sizeof (*has_slot));
    kslot = xmalloc (ninsert * sizeof (*kslot));
    qsort (insert, ninsert, sizeof (*insert), i64_cmp);
    while (ninsert && *insert < 0) {
      ++insert;
      --ninsert;
    }
  }

  if (!nremove && !ninsert) {
    return nrem;
  }

  curs = etype_begin (&G->LVA[v], type);
  prev_loc = curs.loc;

  for (tmp = curs.eb; tmp && type == tmp->etype; tmp = tmp->next) {
    size_t k, endk;
    size_t found_k = STINGER_EDGEBLOCKSIZE;
    size_t found_slot = STINGER_EDGEBLOCKSIZE;
    prev_loc = &tmp->next;
    endk = tmp->high;
    ninsert_remaining = ninsert;

      for (k = 0; k < endk; ++k) {
        const int64_t w = tmp->edges[k].neighbor;
        int64_t off;

        if (w >= 0) {
          /* Removal first.  If the vertex is removed,
             save its slot for reuse. */
          if (nremove) {
            off = find_in_sorted (w, nremove, remove);
            if (off >= 0) {
              int64_t which = stinger_int64_fetch_add (&nslot, 1);
              if (which < ninsert_remaining) {    /* Can be racy. */
                has_slot[which] = tmp;
                kslot[which] = k;
              }
	      remove_edge(G, tmp, k);
              stinger_int64_fetch_add (&nrem, 1);
            }
          }
          if (ninsert) {
            /* Check if this is to be inserted. */
            off = find_in_sorted (w, ninsert, insert);
            while (off >= 0) {
              stinger_int64_fetch_add (&ninsert_remaining, -1);
              insert[off] = ~insert[off];
	      update_edge_data(G, tmp, k, w, (weight ? weight[off] : 1), timestamp);
	      qsort (insert, ninsert, sizeof (*insert), i64_cmp);  /* Must maintain sorted order here. */
	      off = find_in_sorted (w, ninsert, insert);   /* Gotta check if there's another one. */
            }
          }
        } else if (ninsert_remaining) {
          /* Empty slot.  Save. */
          int64_t which = stinger_int64_fetch_add (&nslot, 1);
          if (which < ninsert_remaining) {        /* Can be racy. */
            has_slot[which] = tmp;
            kslot[which] = k;
          }
        }
      }

    MTA ("mta parallel off") if (nslot < ninsert_remaining) {
      /* Gather any trailing slots. */
      MTA ("mta assert nodep")
        for (; endk < STINGER_EDGEBLOCKSIZE; ++endk) {
          int64_t which = stinger_int64_fetch_add (&nslot, 1);
          if (which < ninsert_remaining) {        /* Can be racy. */
            has_slot[which] = tmp;
            kslot[which] = endk;
          }
        }
    }
  }

  /* At this point, know how many need inserted & how many slots
     already are available. */

  if (ninsert_remaining > nslot) {
    /* Need more edge blocks. */
    int64_t neb =
      (ninsert_remaining + STINGER_EDGEBLOCKSIZE - 1) / STINGER_EDGEBLOCKSIZE;
    struct stinger_eb **ebs;
    ebs = xmalloc (neb * sizeof (*ebs));

    new_ebs (ebs, neb, type, v);
    for (int64_t kb = 0; kb < neb - 1; ++kb)
      ebs[kb]->next = ebs[kb + 1];
    ebs[neb - 1]->next = tmp;
    *prev_loc = ebs[0];
    push_ebs (G, neb, ebs);

    MTA ("mta assert nodep")
      for (int64_t kb = 0; kb < neb; ++kb) {
        struct stinger_eb *eb = ebs[kb];
        MTA ("mta assert nodep") MTASTREAMS ()
          for (int64_t k = 0;
               nslot < ninsert_remaining
                 && k <
                 STINGER_EDGEBLOCKSIZE; ++k) {
            int64_t which = stinger_int64_fetch_add (&nslot, 1);
            if (which < ninsert_remaining) {        /* Can be racy. */
              has_slot[which] = eb;
              kslot[which] = k;
            }
          }
      }

    free (ebs);
  }

  if (ninsert_remaining) {
    nslot = 0;
    for (int64_t k = 0; k < ninsert; ++k) {
      const int64_t w = insert[k];
      if (w >= 0) {
        int64_t ks;
        struct stinger_eb *restrict eb;
        int64_t which = stinger_int64_fetch_add (&nslot, 1);

        assert (which < ninsert_remaining);
        ks = kslot[which];
        eb = has_slot[which];

        assert (ks < STINGER_EDGEBLOCKSIZE);
        assert (ks >= 0);
        assert (eb);
        /* Breaking atomicity => assert may break. */
        update_edge_data (G, eb, ks, w,
                          (weight ? weight[k] : 1), timestamp);
      }
    }
  }

  free (kslot);
  free (has_slot);
  return (nrem + ninsert_remaining) > 0;
}

int
stinger_remove_and_insert_edges (struct stinger *G,
                                 int64_t type, int64_t from,
                                 int64_t nremove, int64_t * remove,
                                 int64_t ninsert, int64_t * insert,
                                 int64_t * weight, int64_t timestamp)
{
  int64_t v;
  assert (G);

  v = from;
  assert (v >= 0);
  assert (v < STINGER_MAX_LVERTICES);
#if !defined(NDEBUG)
  for (int64_t k = 0; k < ninsert; ++k) {
    const int64_t w = insert[k];
    //assert (w >= 0);
    assert (w < STINGER_MAX_LVERTICES);
  }
  for (int64_t k = 0; k < nremove; ++k) {
    const int64_t w = remove[k];
    //assert (w >= 0);
    assert (w < STINGER_MAX_LVERTICES);
  }
#endif

  return remove_and_insert_active_edges (G, type, v, nremove, remove,
                                         ninsert, insert, weight, timestamp);
}

/** @brief Initializes an empty STINGER with a graph in CSR format.
 *
 *  Takes an edge list in CSR format, with weight and timestamps, and initializes
 *  an empty STINGER based on the input graph.  All edges being ingested must be
 *  of a single edge type.
 *
 *  @param G STINGER data structure
 *  @param nv Number of vertices
 *  @param etype Edge type
 *  @param off_in Array of length nv containing the adjacency offset for each vertex
 *  @param phy_adj_in Array containing the destination vertex of each edge
 *  @param weight_in Array containing integer weight of each edge
 *  @param ts_in Array containing recent timestamp of edge edge (or NULL)
 *  @param first_ts_in Array containing the first timestamp of each edge (or NULL)
 *  @param single_ts Value for timestamps if either of the above is NULL
 *  @return Void
 */
MTA ("mta parallel default") MTA ("mta inline")
void
stinger_set_initial_edges (struct stinger *G,
                           const size_t nv,
                           const int64_t etype,
                           const int64_t * off_in,
                           const int64_t * phys_adj_in,
                           const int64_t * weight_in,
                           const int64_t * ts_in,
                           const int64_t * first_ts_in,
                           const int64_t single_ts
                           /* if !ts or !first_ts */ )
{
  const int64_t *restrict off = off_in;
  const int64_t *restrict phys_adj = phys_adj_in;
  const int64_t *restrict weight = weight_in;
  const int64_t *restrict ts = ts_in;
  const int64_t *restrict first_ts = first_ts_in;
  struct stinger_vb *restrict LVA;

  size_t nblk_total = 0;
  size_t *restrict blkoff;
  struct stinger_eb **restrict block;

  assert (G);
  LVA = G->LVA;

  //stinger_activate_vtx_range (G, 0, nv, DEFAULT_VTYPE);

  blkoff = xcalloc (nv + 1, sizeof (*blkoff));
  OMP ("omp parallel for")
    for (int64_t v = 0; v < nv; ++v) {
      const int64_t deg = off[v + 1] - off[v];
      blkoff[v + 1] = (deg + STINGER_EDGEBLOCKSIZE - 1) / STINGER_EDGEBLOCKSIZE;
    }

  for (int64_t v = 2; v <= nv; ++v)
    blkoff[v] += blkoff[v - 1];
  nblk_total = blkoff[nv];

  block = xcalloc (nblk_total, sizeof (*block));
  OMP ("omp parallel for")
    MTA ("mta assert nodep") MTASTREAMS ()
    for (int64_t v = 0; v < nv; ++v) {
      const int64_t from = v;
      const int64_t deg = off[v + 1] - off[v];
      if (deg)
        stinger_int64_fetch_add (&LVA[from].outDegree, deg);
    }

  new_blk_ebs (&block[0], G, nv, blkoff, etype);

  /* XXX: AUGH! I cannot find what really is blocking parallelization. */
  MTA ("mta assert parallel")
    MTA ("mta dynamic schedule")
    OMP ("omp parallel for")
    MTA ("mta assert noalias *block")
    MTA ("mta assert nodep *block")
    MTA ("mta assert noalias *G")
    MTA ("mta assert nodep *G")
    MTA ("mta assert nodep *phys_adj")
    MTA ("mta assert nodep *LVA")
    MTA ("mta assert nodep *blkoff")
    for (int64_t v = 0; v < nv; ++v) {
      const size_t nextoff = off[v + 1];
      size_t kgraph = off[v];
      MTA ("mta assert may reorder kgraph") int64_t from;

      from = v;

      // Need to assert this loop is not to be parallelized!
      MTA ("mta assert nodep *block")
        MTA ("mta assert nodep *LVA")
        for (size_t kblk = blkoff[v]; kblk < blkoff[v + 1]; ++kblk) {
          size_t n_to_copy, voff;
          struct stinger_edge *restrict edge;
          struct stinger_eb *restrict eb;
          int64_t tslb = INT64_MAX, tsub = 0;

          //voff = stinger_size_fetch_add (&kgraph, STINGER_EDGEBLOCKSIZE);
          {
            voff = kgraph;
            kgraph += STINGER_EDGEBLOCKSIZE;
          }

          n_to_copy = STINGER_EDGEBLOCKSIZE;
          if (voff + n_to_copy >= nextoff)
            n_to_copy = nextoff - voff;

          eb = block[kblk];
          edge = &eb->edges[0];

          /* XXX: remove the next two asserts once the outer is unlocked. */
          MTA ("mta assert nodep")
            MTASTREAMS ()MTA ("mta assert nodep *phys_adj")
            MTA ("mta assert nodep *edge")
            MTA ("mta assert nodep *LVA")
            for (size_t i = 0; i < n_to_copy; ++i) {
              const int64_t to = phys_adj[voff + i];
#if defined(__MTA__)
              int_fetch_add (&LVA[to].inDegree, 1);
#else
              // Ugh. The MTA compiler can't cope with the inlining.
              stinger_int64_fetch_add (&LVA[to].inDegree, 1);
#endif
              /* XXX: The next statements block parallelization
                 of the outer loop. */
              edge[i].neighbor = to;
              edge[i].weight = weight[voff + i];
              edge[i].timeRecent = ts ? ts[voff + i] : single_ts;
              edge[i].timeFirst = first_ts ? first_ts[voff + i] : single_ts;
              //assert (edge[i].timeRecent >= edge[i].timeFirst);
            }

          if (ts || first_ts) {
            /* XXX: remove the next two asserts once the outer is unlocked. */
            MTA ("mta assert nodep")
              MTASTREAMS ()MTA ("mta assert nodep *edge")
              for (size_t i = 0; i < n_to_copy; ++i) {
                if (edge[i].timeFirst < tslb)
                  tslb = edge[i].timeFirst;
                if (edge[i].timeRecent > tsub)
                  tsub = edge[i].timeRecent;
              }
          } else
            tslb = tsub = single_ts;

          eb->smallStamp = tslb;
          eb->largeStamp = tsub;
          eb->numEdges = n_to_copy;
          eb->high = n_to_copy;
        }


      /* At this point, block[blkoff[v]] is the head of a linked
         list holding all the blocks of edges of EType from vertex
         v.  Insert into the graph.  */

      if (blkoff[v] != blkoff[v + 1])
        LVA[from].edges = block[blkoff[v]];
      else
        LVA[from].edges = NULL;
    }

  /* Insert into the edge type array */
  push_ebs (G, nblk_total, block);

  free (block);
  free (blkoff);
}

/** @brief Copy typed incoming adjacencies of a vertex into a buffer
 *
 *  For a given edge type, adjacencies of the specified vertex are copied into
 *  the user-provided buffer up to the length of the buffer.  These are the
 *  incoming edges for which a vertex is a destination.  Note that this operation
 *  may be very expensive on most platforms.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param v Source vertex ID
 *  @param outlen Number of adjacencies copied
 *  @param out Buffer to hold adjacencies
 *  @param max_outlen Length of out[] and recent[]
 *  @return Void
 */
MTA ("mta inline")
void
stinger_gather_typed_predecessors (const struct stinger *G,
				   int64_t type,
				   int64_t v,
				   size_t * outlen,
                                   int64_t * out,
                                   size_t max_outlen)
{
  size_t kout = 0;

  assert (G);

  STINGER_PARALLEL_FORALL_EDGES_BEGIN(G, type) {
    const int64_t u = STINGER_EDGE_SOURCE;
    const int64_t v = STINGER_EDGE_DEST;
    if (v >= 0) {
      size_t where = stinger_size_fetch_add (&kout, 1);
      if (where < max_outlen) {
        out[where] = u;
      }
    }
  } STINGER_PARALLEL_FORALL_EDGES_END();

  *outlen = kout;               /* May be longer than max_outlen. */
}

/** @brief Copy typed adjacencies of a vertex into a buffer with recent timestamps
 *
 *  For a given edge type, adjacencies of the specified vertex are copied into
 *  the user-provided buffer up to the length of the buffer.  The recent timestamp
 *  is also returned in a separate buffer.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param v Source vertex ID
 *  @param outlen Number of adjacencies copied
 *  @param out Buffer to hold adjacencies
 *  @param recent Buffer to hold recent timestamps
 *  @param max_outlen Length of out[] and recent[]
 *  @return Void
 */
MTA ("mta inline")
void
stinger_gather_typed_successors_recent (const struct stinger *G,
                                          int64_t type, int64_t v,
                                          size_t * outlen,
                                          int64_t * out,
                                          int64_t * recent,
                                          size_t max_outlen)
{
  size_t kout = 0;

  assert (G);

  STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_BEGIN(G, type, v) {
    const int64_t n = STINGER_EDGE_DEST;
    const int64_t ts = STINGER_EDGE_TIME_RECENT;
    if (n >= 0) {
      size_t where = stinger_size_fetch_add (&kout, 1);
      if (where < max_outlen) {
        out[where] = n;
        recent[where] = ts;
      }
    }
  } STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_END();

  *outlen = kout;               /* May be longer than max_outlen. */
}

/** @brief Copy typed adjacencies of a vertex into a buffer with edge weights
 *
 *  For a given edge type, adjacencies of the specified vertex are copied into
 *  the user-provided buffer up to the length of the buffer.  The edge weight
 *  is also returned in a separate buffer.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param v Source vertex ID
 *  @param outlen Number of adjacencies copied
 *  @param out Buffer to hold adjacencies
 *  @param ewgts Buffer to hold edge weights
 *  @param max_outlen Length of out[] and recent[]
 *  @return Void
 */
MTA ("mta inline")
void
stinger_gather_typed_successors_weighted (const struct stinger *G,
                                          int64_t type, int64_t v,
                                          size_t * outlen,
                                          int64_t * out,
                                          int64_t * ewgts,
                                          size_t max_outlen)
{
  size_t kout = 0;

  assert (G);

  STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_BEGIN(G, type, v) {
    const int64_t n = STINGER_EDGE_DEST;
    const int64_t ew = STINGER_EDGE_WEIGHT;
    if (n >= 0) {
      size_t where = stinger_size_fetch_add (&kout, 1);
      if (where < max_outlen) {
        out[where] = n;
        ewgts[where] = ew;
      }
    }
  } STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_END();

  *outlen = kout;               /* May be longer than max_outlen. */
}

/** @brief Copy typed adjacencies of a vertex into a buffer
 *
 *  For a given edge type, adjacencies of the specified vertex are copied into
 *  the user-provided buffer up to the length of the buffer.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param v Source vertex ID
 *  @param outlen Number of adjacencies copied
 *  @param out Buffer to hold adjacencies
 *  @param max_outlen Length of out[] and recent[]
 *  @return Void
 */
MTA ("mta inline")
void
stinger_gather_typed_successors (const struct stinger *G, int64_t type,
                                 int64_t v, size_t * outlen,
                                 int64_t * out, size_t max_outlen)
{
  size_t kout = 0;

  assert (G);

  if (v < 0) {
    *outlen = 0;
    return;
  }

  STINGER_FORALL_EB_BEGIN (G, v, eb) {
    if (stinger_eb_type (eb) > type)
      break;
    else if (stinger_eb_type (eb) == type) {
      const size_t eblen = stinger_eb_high (eb);
      OMP ("omp parallel for")
        MTA ("mta assert nodep")
        MTASTREAMS ()
        for (size_t ek = 0; ek < eblen; ++ek) {
          if (!stinger_eb_is_blank (eb, ek)) {
            const int64_t n = stinger_eb_adjvtx (eb, ek);
            size_t where = stinger_size_fetch_add (&kout, 1);
            if (where < max_outlen)
              out[where] = n;
          }
        }
    }
  }
  STINGER_FORALL_EB_END ();

  *outlen = kout;               /* May be longer than max_outlen. */
}

/** @brief Copy typed adjacencies of a vertex into a buffer
 *
 *  For a given edge type, adjacencies of the specified vertex are copied into
 *  the user-provided buffer up to the length of the buffer.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param v Source vertex ID
 *  @param outlen Number of adjacencies copied
 *  @param out Buffer to hold adjacencies
 *  @param max_outlen Length of out[] and recent[]
 *  @return Void
 */
void
stinger_gather_typed_successors_serial (const struct stinger *G, int64_t type,
                                        int64_t v, size_t * outlen,
                                        int64_t * out, size_t max_outlen)
{
  size_t kout = 0;

  assert (G);

  if (v < 0) {
    *outlen = 0;
    return;
  }

  STINGER_FORALL_EB_BEGIN (G, v, eb) {
    if (stinger_eb_type (eb) > type)
      break;
    else if (stinger_eb_type (eb) == type) {
      const size_t eblen = stinger_eb_high (eb);
      for (size_t ek = 0; ek < eblen; ++ek) {
        if (!stinger_eb_is_blank (eb, ek)) {
          const int64_t n = stinger_eb_adjvtx (eb, ek);
          size_t where = stinger_size_fetch_add (&kout, 1);
          if (where < max_outlen)
            out[where] = n;
        }
      }
    }
  }
  STINGER_FORALL_EB_END ();

  *outlen = kout;               /* May be longer than max_outlen. */
}

/** @brief Determines if a vertex has an edge of a given type
 *
 *  Searches the adjacencies of a specified vertex for an edge of the given type.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @return 1 if found; 0 otherwise
 */
MTA ("mta inline")
int
stinger_has_typed_successor (const struct stinger *G,
                             int64_t type, int64_t from, int64_t to)
{
  STINGERASSERTS ();

  int rtn = 0;

	STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_BEGIN(G,type,from) {
		if (STINGER_EDGE_DEST == to) {
      rtn = 1;
    }
	} STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_END();
  return rtn;
}

/** @brief Get the weight of a given edge.
 *
 *  Finds a specified edge of a given type by source and destination vertex ID
 *  and returns the current edge weight.  Remember, edges may have different
 *  weights in different directions.
 *
 *  @param G The STINGER data structure
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @param type Edge type
 *  @return Edge weight
 */
MTA ("mta inline")
int64_t
stinger_edgeweight (const struct stinger * G,
                    int64_t from, int64_t to, int64_t type)
{
  STINGERASSERTS ();

  int rtn = 0;

	STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_BEGIN(G,type,from) {
		if (STINGER_EDGE_DEST == to) {
      rtn = STINGER_EDGE_WEIGHT;
    }
	} STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_END();
  return rtn;
}

/** @brief Set the weight of a given edge.
 *
 *  Finds a specified edge of a given type by source and destination vertex ID
 *  and sets the current edge weight.  Remember, edges may have different
 *  weights in different directions.
 *
 *  @param G The STINGER data structure
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @param type Edge type
 *  @param weight Edge weight to set
 *  @return 1 on success; 0 otherwise
 */
MTA ("mta inline")
int
stinger_set_edgeweight (struct stinger *G,
                        int64_t from, int64_t to, int64_t type,
                        int64_t weight)
{
  STINGERASSERTS ();

  int rtn = 0;

	STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_BEGIN(G,type,from) {
		if (STINGER_EDGE_DEST == to) {
      STINGER_EDGE_WEIGHT = weight;
      rtn = 1;
    }
	} STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_END();
  return rtn;
}

/** @brief Get the first timestamp of a given edge.
 *
 *  Finds a specified edge of a given type by source and destination vertex ID
 *  and returns the first timestamp.  Remember, edges may have different
 *  timestamps in different directions.
 *
 *  @param G The STINGER data structure
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @param type Edge type
 *  @return First timestamp of edge; -1 if edge does not exist
 */
MTA ("mta inline")
int64_t
stinger_edge_timestamp_first (const struct stinger * G,
                              int64_t from, int64_t to, int64_t type)
{
  STINGERASSERTS ();

  int rtn = -1;

	STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_BEGIN(G,type,from) {
		if (STINGER_EDGE_DEST == to) {
      rtn = STINGER_EDGE_TIME_FIRST;
    }
	} STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_END();
  return rtn;
}

/** @brief Get the recent timestamp of a given edge.
 *
 *  Finds a specified edge of a given type by source and destination vertex ID
 *  and returns the recent timestamp.  Remember, edges may have different
 *  timestamps in different directions.
 *
 *  @param G The STINGER data structure
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @param type Edge type
 *  @return Recent timestamp of edge; -1 if edge does not exist
 */
MTA ("mta inline")
int64_t
stinger_edge_timestamp_recent (const struct stinger * G,
                               int64_t from, int64_t to, int64_t type)
{
  STINGERASSERTS ();

  int rtn = -1;

	STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_BEGIN(G,type,from) {
		if (STINGER_EDGE_DEST == to) {
      rtn = STINGER_EDGE_TIME_RECENT;
    }
	} STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_END();
  return rtn;
}

/** @brief Update the recent timestamp of an edge
 *
 *  Finds a specified edge of a given type by source and destination vertex ID
 *  and updates the recent timestamp to the specified value.
 *
 *  @param G The STINGER data structure
 *  @param from Source vertex ID
 *  @param to Destination vertex ID
 *  @param type Edge type
 *  @param timestamp Timestamp to set recent
 *  @return 1 on success; 0 if failure
 */
MTA ("mta inline")
int
stinger_edge_touch (struct stinger *G,
                    int64_t from, int64_t to, int64_t type,
                    int64_t timestamp)
{
  STINGERASSERTS ();

  int rtn = 0;

	STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_BEGIN(G,type,from) {
		if (STINGER_EDGE_DEST == to) {
      STINGER_EDGE_TIME_RECENT = timestamp;
      rtn = 1;
    }
	} STINGER_PARALLEL_FORALL_EDGES_OF_TYPE_OF_VTX_END();
  return rtn;
}

MTA ("mta inline") int64_t
stinger_count_outdeg (struct stinger * G, int64_t v)
{
  int64_t nactive_edges = 0, neb = 0;

  STINGER_FORALL_EB_BEGIN (G, v, eb) {
    const size_t eblen = stinger_eb_high (eb);
    OMP ("omp parallel for")
      MTA ("mta assert nodep")
      MTASTREAMS ()
      for (size_t ek = 0; ek < eblen; ++ek) {
        if (!stinger_eb_is_blank (eb, ek))
          stinger_int64_fetch_add (&nactive_edges, 1);
      }
  }
  STINGER_FORALL_EB_END ();

  return nactive_edges;
}

MTA ("mta inline")
static int i2cmp (const void *va, const void *vb)
{
  const int64_t *a = va;
  const int64_t *b = vb;
  if (a[0] < b[0])
    return -1;
  if (a[0] > b[0])
    return 1;
  if (a[1] < b[1])
    return -1;
  if (a[1] > b[1])
    return 1;
  return 0;
}

/**
* @brief Sorts a batch of edge insertions and deletions.
* 
* Takes an array of edge insertions and deletions and sorts them.  The array is
* packed with <source, destination> pairs, such that actions[2*i] = source vertex
* ID and actions[2*i+1] = destination vertex ID.  Bit-wise complement the source
* and destination vertex IDs to indicate a deletion.  This function will create
* an undirected actions list (i.e. create the reverse edge action).
*
* The output is similarly packed, sorted by source vertex ID first, then by
* destination vertex ID such that deletions are contiguous before insertions.
* For vertex i, deletions incident on i start at deloff[i] and end at insoff[i].
* Insertions incident on i start at insoff[i] and end at deloff[i+1].
*
* @param nactions Number of edge actions
* @param actions The packed array of edge insertions and deletions
* @param insoff For each incident vertex in the batch, the offset into act[] of
*   the first edge insertion
* @param deloff For each incident vertex in the batch, the offset into act[] of
*   the first edge deletion
* @param act Sorted array of edge insertions and deletions
*
* @return The number of incident vertices in the batch (the size of insoff[] and deloff[])
*/
int64_t
stinger_sort_actions (int64_t nactions, int64_t * actions,
                      int64_t * insoff, int64_t * deloff, int64_t * act)
{
  int64_t head = 0;
  int64_t actlen;
  int64_t n;
  int64_t *actk = xmalloc ((2 * 2 * nactions + 1) * sizeof (*actk));

  OMP("omp parallel") {
  /* Copy & make i positive if necessary.
     Negative j still implies deletion. */
  MTA ("mta assert nodep")
    MTA ("mta block schedule")
    OMP ("omp for")
    for (int64_t k = 0; k < nactions; k++) {
      const int64_t i = actions[2 * k];
      const int64_t j = actions[2 * k + 1];

      if (i != j) {
        int64_t where = stinger_int64_fetch_add (&head, 2);
        assert (where < 2 * nactions);

        act[2 * where] = (i < 0 ? -i - 1 : i);
        act[2 * where + 1] = j;
        act[2 * (where + 1)] = (j < 0 ? -j - 1 : j);
        act[2 * (where + 1) + 1] = i;
      }
    }

  OMP("omp single") {
  actlen = head;
  //qsort (act, actlen, 2 * sizeof (act[0]), i2cmp); 
  radix_sort_pairs (act, actlen<<1, 6);

  actk[0] = 0;
  n = 1;
  }

  /* Find local indices... */
  MTA ("mta assert nodep")
    OMP ("omp for")
    for (int64_t k = 1; k < actlen; k++) {
      if (act[2 * k] != act[2 * (k - 1)]) {
        stinger_int64_fetch_add (&n, 1);
        actk[k] = 1;
      } else
        actk[k] = 0;
    }

  OMP("omp single")
  for (int64_t k = 1; k < actlen; k++) {
    actk[k] += actk[k - 1];
  }
  //XXX assert(actk[actlen-1] == n-1);

  MTA ("mta assert nodep")
    OMP ("omp for")
    for (int64_t k = 0; k <= n; k++)
      deloff[k] = 0;

  OMP ("omp for")
    MTA ("mta assert nodep")
    for (int64_t k = 0; k < actlen; k++)
      stinger_int64_fetch_add (&deloff[actk[k] + 1], 1);

  OMP("omp single")
  for (int64_t k = 1; k <= n; k++)
    deloff[k] += deloff[k - 1];
  assert (deloff[n] == actlen);

  MTA ("mta assert nodep")
    MTA ("mta loop norestructure")
    OMP ("omp for")
    for (int64_t k = 0; k < n; k++) {
      int off;
      const int endoff = deloff[k + 1];

      for (off = deloff[k]; off < endoff; off++)
        if (act[2 * off + 1] >= 0)
          break;
      insoff[k] = off;
      assert (insoff[k] == deloff[k + 1]
              || act[2 * insoff[k]] == act[2 * deloff[k]]);
      for (off = deloff[k]; off < insoff[k]; off++) {
        assert (act[2 * off + 1] < 0);
        act[2 * off + 1] = -act[2 * off + 1] - 1;
      }
    }
  }

  free (actk);

  return n;                     /* number of unique vertices, elements in insoff[], deloff[] */
}

/** @brief Process edge removals and insertions as a batch.
 *
 *  Takes its input from stinger_sort_actions().  Takes a sorted batch of edge
 *  insertions and removals and processes them in parallel in the data structure.
 *
 *  @param G The STINGER data structure
 *  @param type Edge type
 *  @param timestamp The current timestamp for this batch
 *  @param n Number of incident vertices in the batch
 *  @param insoff For each incident vertex, the offset into the actions array of insertions
 *  @param deloff For each incident vertex, the offset into the actions array of deletions
 *  @param act The sorted actions array
 *  @return The number of incident vertices
 */
int64_t
stinger_remove_and_insert_batch (struct stinger * G, int64_t type,
                                 int64_t timestamp, int64_t n,
                                 int64_t * insoff, int64_t * deloff,
                                 int64_t * act)
{

  /* Separate each vertex's batch into deletions and insertions */
  OMP ("omp parallel for")
    MTA ("mta assert nodep")
    for (int k = 0; k < n; k++) {
      const int64_t i = act[2 * deloff[k]];
      const int64_t nrem = insoff[k] - deloff[k];
      const int64_t nins = deloff[k + 1] - insoff[k];
      int64_t *restrict torem, *restrict toins;
      torem = (nrem + nins ? xmalloc ((nrem + nins) * sizeof (*torem)) : NULL);
      toins = (torem ? &torem[nrem] : NULL);

      MTA ("mta assert nodep")
        for (int k2 = 0; k2 < nrem; ++k2) {
          torem[k2] = act[2 * (deloff[k] + k2) + 1];
          assert (act[2 * (deloff[k] + k2)] == i);
        }

      MTA ("mta assert nodep")
        for (int k2 = 0; k2 < nins; k2++) {
          toins[k2] = act[2 * (insoff[k] + k2) + 1];
          assert (act[2 * (insoff[k] + k2)] == i);
        }

      /* Do the update in the data structure */
      stinger_remove_and_insert_edges (G, type, i, nrem, torem, nins, toins,
                                       NULL, timestamp);
      if (torem != NULL)
	free (torem);
    }

  return n;
}

/** @brief Removes all the edges in the graph of a given type.
 *
 *  Traverses all edge blocks of a particular type in parallel and erases all
 *  edges of the specified type.  Blocks are available immediately for reuse.
 *
 *  @param G The STINGER data structure
 *  @param type The edge type of edges to delete
 *  @return Void
 */
void
stinger_remove_all_edges_of_type (struct stinger *G, int64_t type)
{
  MTA("mta assert nodep")
  for (uint64_t p = 0; p < G->ETA[type].high; p++) {
    struct stinger_eb *current_eb = G->ETA[type].blocks[p];
    for (uint64_t i = 0; i < stinger_eb_high (current_eb); i++) {
      current_eb->edges[i].neighbor = -1;
    }
    current_eb->high = 0;
    current_eb->numEdges = 0;
    current_eb->smallStamp = -1;
    current_eb->largeStamp = -1;
  }
}

/** @brief Checkpoint a STINGER data structure to disk.
 *
 *  Creates a directory if needed and stores the graph as individual adjacency files in directory
 *  for parallelism / no need for extra buffers.
 *  Format:
 *  endian_check -> 64-bit endianness file
 *  nv -> 64-bit max vertex ID
 *  0 .. maxVtx -> 64-bit outdegree followed by adjacencies of vertex 0 stored as 5-tuples 
 *     (type, dest, weight, timefirst, timerecent)
 *
 *  @param S The STINGER data structure
 *  @param maxVtx Largest logical vertex ID + 1
 *  @param stingerfiledir String containing the name of the directory to store the checkpoint files in
 *  @return 0 on success; -1 if error
 */
int
stinger_save_to_file (struct stinger * S, uint64_t maxVtx, const char * stingerfiledir)
{
   /* TODO: 
   * - XMT testing / code
   * - Performance / size tweaks (this is a rough first run)
   */

#if !defined(__MTA__)
  if(stinger_consistency_check(S, maxVtx)) {
    fprintf(stderr, "WARNING: Stinger is inconsistent before writing to disk.\n");
  }

  mkdir(stingerfiledir, S_IRWXU | S_IRWXG | S_IROTH);

  if(chdir(stingerfiledir)) {
    fprintf(stderr, "ERROR: Directory creating / cd Failed.\n");
    return -1;
  }

  struct dirent **names;
  int64_t n = scandir(".", &names, 0, alphasort);
  if(n >= 0) {
    while(n--) {
      if(names[n]->d_type == DT_REG)
	unlink(names[n]->d_name);
      free(names[n]);
    }
  }
  free(names);

  {
    const int64_t endian_check = 0x1234ABCD;
    FILE * f_endianness = fopen("endian_check", "w"); 
    if(!f_endianness) {
      fprintf(stderr, "ERROR: Creating endianness failed.\n");
      return -1;
    }
    if(1 != fwrite(&endian_check, sizeof(int64_t), 1, f_endianness)) {
      fprintf(stderr, "ERROR: Writing endianness failed.\n");
      return -1;
    }
    fclose(f_endianness);
  }

  {
    FILE * f_nv = fopen("nv", "w"); 
    if(!f_nv || 1 != fwrite(&maxVtx, sizeof(int64_t), 1, f_nv)) {
      fprintf(stderr, "ERROR: Writing NV failed.\n");
      return -1;
    }
    fclose(f_nv);
  }

  uint64_t error = 0;

  OMP("omp parallel for")
  for(uint64_t i = 0; i < maxVtx; i++) {
    uint64_t outDegree = stinger_outdegree(S, i);
    if(outDegree) {
      char filename[256];
      sprintf(filename, "%lu", i);
      FILE * f_curvtx = fopen(filename, "w");
      if(!f_curvtx || 1 != fwrite(&outDegree, sizeof(uint64_t), 1, f_curvtx)) {
	fprintf(stderr, "ERROR: Writing %lu outdegree failed.\n", i);
	error = -1;
      } else {
	uint64_t value = stinger_vweight(S, i);
	if(1 != fwrite(&value, sizeof(uint64_t), 1, f_curvtx)) {
	  fprintf(stderr, "ERROR: Writing %lu vweight failed.\n", i);
	  error = -1;
	}
	value = stinger_vtype(S, i);
	if(1 != fwrite(&value, sizeof(uint64_t), 1, f_curvtx)) {
	  fprintf(stderr, "ERROR: Writing %lu vtype failed.\n", i);
	  error = -1;
	}
	STINGER_FORALL_EDGES_OF_VTX_BEGIN(S, i) {
	  if(1 != fwrite(&(STINGER_EDGE_TYPE), sizeof(uint64_t), 1, f_curvtx)) {
	    fprintf(stderr, "ERROR: Writing %lu type failed.\n", i);
	    error = -1;
	  }
	  if(1 != fwrite(&(STINGER_EDGE_DEST), sizeof(uint64_t), 1, f_curvtx)) {
	    fprintf(stderr, "ERROR: Writing %lu dest failed.\n", i);
	    error = -1;
	  }
	  if(1 != fwrite(&(STINGER_EDGE_WEIGHT), sizeof(uint64_t), 1, f_curvtx)) {
	    fprintf(stderr, "ERROR: Writing %lu weight failed.\n", i);
	    error = -1;
	  }
	  if(1 != fwrite(&(STINGER_EDGE_TIME_FIRST), sizeof(uint64_t), 1, f_curvtx)) {
	    fprintf(stderr, "ERROR: Writing %lu timefirst failed.\n", i);
	    error = -1;
	  }
	  if(1 != fwrite(&(STINGER_EDGE_TIME_RECENT), sizeof(uint64_t), 1, f_curvtx)) {
	    fprintf(stderr, "ERROR: Writing %lu timerecent failed.\n", i);
	    error = -1;
	  }
	} STINGER_FORALL_EDGES_OF_VTX_END();
	fclose(f_curvtx);
      }
    }
  }
  if(chdir("..")) {
    fprintf(stderr, "ERROR: Returning to parent.\n");
    error = -1;
  }

  return error;
#endif
}

/** @brief Restores a STINGER checkpoint from disk.
 *
 *  Reads STINGER checkpoint files from disk and reconstructs the STINGER data structure
 *  according to their contents.
 *
 *  @param stingerfiledir Directory name containing the STINGER checkpoint files
 *  @param S STINGER data structure to restore into
 *  @param maxVtx Largest logical vertex ID + 1
 *  @return 0 on success; -1 if error 
 */
int
stinger_open_from_file (const char * stingerfiledir, struct stinger ** S, uint64_t * maxVtx)
{

#if !defined(__MTA__)
  *S = NULL;
  if(chdir(stingerfiledir)) {
    fprintf(stderr, "ERROR: Changing to directory.\n");
    *S = NULL;
    return -1;
  }
  int64_t endian_check = 0x1234ABCD;
  {
    FILE * f_endianness = fopen("endian_check", "r");
    int64_t endian_input;
    if(!f_endianness || 1 != fread(&endian_input, sizeof(int64_t), 1, f_endianness)) {
      fprintf(stderr, "ERROR: Can't read endianness file.\n");
      *S = NULL;
      return -1;
    }
    endian_check = endian_check != endian_input;
  }

  {
    FILE * f_nv = fopen("nv", "r");
    if(!f_nv || 1 != fread(maxVtx, sizeof(int64_t), 1, f_nv)) {
      fprintf(stderr, "ERROR: Can't read maxvtx file.\n");
      *S = NULL;
      return -1;
    }
    if(endian_check) {
      *maxVtx = bs64(*maxVtx);
    }
  }

  *S = stinger_new();

  int error = 0;
  OMP("omp parallel for")
  for(uint64_t i = 0; i < *maxVtx; i++) {
    char filename[256];
    sprintf(filename, "%lu", i);
    FILE * f_curvtx = fopen(filename, "r");
    if(f_curvtx) {
      uint64_t outDegree;
      if(1 != fread(&outDegree, sizeof(int64_t), 1, f_curvtx)) {
	fprintf(stderr, "ERROR: Can't read %lu file.\n", i);
	error = -1;
      }
      if(endian_check) {
	outDegree = bs64(outDegree);
      }
      uint64_t value = 0;
      if(1 != fread(&value, sizeof(int64_t), 1, f_curvtx)) {
	fprintf(stderr, "ERROR: Can't read %lu vweight file.\n", i);
	error = -1;
      }
      if(endian_check) {
	value = bs64(value);
      }
      stinger_set_vweight(*S, i, value);
      if(1 != fread(&value, sizeof(int64_t), 1, f_curvtx)) {
	fprintf(stderr, "ERROR: Can't read %lu vtype file.\n", i);
	error = -1;
      }
      if(endian_check) {
	value = bs64(value);
      }
      stinger_set_vtype(*S, i, value);
      struct stinger_eb * ebstack; 
      uint64_t stackSize = (outDegree + STINGER_EDGEBLOCKSIZE - 1) / STINGER_EDGEBLOCKSIZE;
      uint64_t stackTop = 0;
      int64_t type, dest, weight, tfirst, trecent;
      new_ebs(&ebstack, stackSize, 0, i);
      int read;
      (*S)->LVA[i].edges = ebstack;
      struct stinger_eb * curblock = ebstack;
      while((read = fread(&type, sizeof(uint64_t), 1, f_curvtx)) != 0) {
	if(curblock->high == STINGER_EDGEBLOCKSIZE) {
	  curblock++;
	  stackTop++;
	  if(stackTop == stackSize) {
	    push_ebs(*S, stackSize, &ebstack);
	    stackSize = 1;
	    stackTop = 0;
	    curblock = ebstack;
	    new_ebs(&ebstack, stackSize, 0, i);
	  }
	}

	if(endian_check) {
	  type = bs64(type);
	}
	// handle type
	if(!curblock->numEdges) {
	  curblock->etype = type;
	} else {
	  if(curblock->etype != type) {
	    curblock++;
	    stackTop++;
	    if(stackTop == stackSize) {
	      push_ebs(*S, stackSize, &ebstack);
	      stackSize = 1;
	      stackTop = 0;
	      new_ebs(&ebstack, stackSize, 0, i);
	      curblock = ebstack;
	    } else {
	      push_ebs(*S, stackTop, &ebstack);
	      ebstack += stackTop;
	      stackSize -= stackTop;
	      stackTop = 0;
	      curblock = ebstack;
	    }
	    curblock->etype = type;
	  }
	}

	if(1 != fread(&dest, sizeof(uint64_t), 1, f_curvtx)) {
	  fprintf(stderr, "ERROR: Can't read %lu. Unexpected end\n", i);
	  error = -1;
	}
	if(endian_check) {
	  dest = bs64(dest);
	}
	if(1 != fread(&weight, sizeof(uint64_t), 1, f_curvtx)) {
	  fprintf(stderr, "ERROR: Can't read %lu. Unexpected end\n", i);
	  error = -1;
	}
	if(endian_check) {
	  weight = bs64(weight);
	}
	if(1 != fread(&tfirst, sizeof(uint64_t), 1, f_curvtx)) {
	  fprintf(stderr, "ERROR: Can't read %lu. Unexpected end\n", i);
	  error = -1;
	}
	if(endian_check) {
	  tfirst= bs64(tfirst);
	}
	if(1 != fread(&trecent, sizeof(uint64_t), 1, f_curvtx)) {
	  fprintf(stderr, "ERROR: Can't read %lu. Unexpected end\n", i);
	  error = -1;
	}
	if(endian_check) {
	  trecent = bs64(trecent);
	}
	stinger_int64_fetch_add(&((*S)->LVA[i].outDegree), 1);
	stinger_int64_fetch_add(&((*S)->LVA[dest].inDegree), 1);
	curblock->edges[curblock->high].neighbor = dest;
	curblock->edges[curblock->high].weight = weight;
	curblock->edges[curblock->high].timeFirst = tfirst;
	curblock->edges[curblock->high].timeRecent = trecent;
	if(curblock->smallStamp > tfirst)
	  curblock->smallStamp = tfirst;
	if(curblock->largeStamp < trecent)
	  curblock->largeStamp = trecent;
	curblock->numEdges++;
	curblock->high++;

      }
      push_ebs(*S, stackSize, &ebstack);
      fclose(f_curvtx);
    }
  }

  if(chdir("..")) {
    fprintf(stderr, "ERROR: Returning to parent.\n");
    error = -1;
  }
  return error;
  if(error)
    return error;
  else
    return stinger_consistency_check(*S, *maxVtx);

#endif
}

